﻿(function () {
    var ns = namespace("Entities");

    var Entity = Class(function (out) {
        out.constructor = function Entity(map, id, tileX, tileY) {
            this.map = map;
            this._hat = -1;
            this.id = id;
            this.speed = 1.0;
            this.tileX = tileX;
            this.tileY = tileY;
            this.offsetX = 0.0;
            this.offsetY = 0.0;
            this.color = Color.White;
            this.texture = null;
            this.moving = false;
            this.direction = Constants.DirNone;
            this.facing = Constants.DirDown;
            this.animStart = 0;
            this.keepMoving = false;
            this.obstacle = true;
            this.hatTexture = null;
        }

        out.update = function update() {
            if (this.moving && this.direction == Constants.DirNone) {
                this.startMoving(this.time);
            }

            if (this.direction != Constants.DirNone) {
                this.setMoveOffset(this.time);
            }

            if (this.animations) {
                var frame = this.animations.getFrame(this.curAnim || "stand", this.facing, this.time - this.animStart);
                this.texture = frame.texture;
                if (frame.animName != this.curAnim) {
                    this.switchAnimation(frame.animName);
                }
                if (this.hat > -1) {
                    this.hatTexture = this.hatAnims[this.hat].getFrame(this.curAnim || "stand", this.facing, this.time - this.animStart).texture;
                }
            }
        }

        out.switchAnimation = function switchAnimation(anim) {
            var animation = this.animations.getAnimation(this.curAnim || "stand");
            if (animation.loops && anim == this.curAnim) {
                return;
            }
            this.curAnim = anim;
            if (!animation.keepTiming) {
                this.animStart = this.time;
            }
        }

        out.setMoveOffset = function setMoveOffset() {
            var alpha = Calc.Clamp((this.time - this.moveStartTime) / this.moveDuration);

            if (alpha >= 1.0) {
                if (this.direction == Constants.DirLeft) {
                    this.tileX--;
                }
                else if (this.direction == Constants.DirRight) {
                    this.tileX++;
                }
                else if (this.direction == Constants.DirUp) {
                    this.tileY--;
                }
                else if (this.direction == Constants.DirDown) {
                    this.tileY++;
                }

                this.offsetX = 0;
                this.offsetY = 0;
                this.direction = Constants.DirNone;
                return;
            }

            if (this.direction == Constants.DirLeft) {
                this.offsetX = Calc.Lerp(0.0, -1.0, alpha);
            }
            else if (this.direction == Constants.DirRight) {
                this.offsetX = Calc.Lerp(0.0, 1.0, alpha);
            }
            else if (this.direction == Constants.DirUp) {
                this.offsetY = Calc.Lerp(0.0, -1.0, alpha);
            }
            else if (this.direction == Constants.DirDown) {
                this.offsetY = Calc.Lerp(0.0, 1.0, alpha);
            }
        }

        out.startMoving = function startMoving() {
            if (this.targetX == this.tileX && this.targetY == this.tileY) {
                this.stopMoving();
                return;
            }

            this.switchAnimation("walk");
            this.moveStartTime = this.time;
            this.moveDuration = Constants.MoveDuration * (1.0 / this.speed);

            var toX;
            var toY;

            if (this.targetX < this.tileX) {
                this.direction = Constants.DirLeft;
                toX = this.tileX - 1;
                toY = this.tileY;
            }
            else if (this.targetY < this.tileY) {
                this.direction = Constants.DirUp;
                toX = this.tileX;
                toY = this.tileY - 1;
            }
            else if (this.targetX > this.tileX) {
                this.direction = Constants.DirRight;
                toX = this.tileX + 1;
                toY = this.tileY;
            }
            else if (this.targetY > this.tileY) {
                this.direction = Constants.DirDown;
                toX = this.tileX;
                toY = this.tileY + 1;
            }

            this.facing = this.direction;

            if (!this.map.canMoveTo(this, toX, toY)) {
                this.stopMoving();
                return;
            }
        }

        out.stopMoving = function stopMoving() {
            this.moving = false;
            this.direction = Constants.DirNone;
            if (!this.keepMoving) {
                this.switchAnimation("stand");
            }
            if (this.onStoppedMoving) {
                this.onStoppedMoving(this);
            }
        }

        out.moveTo = function moveTo(x, y) {
            this.moving = true;
            this.targetX = x;
            this.targetY = y;
        }

        out.move = function move(direction, count) {
            if (typeof count === "undefined") {
                count = 1;
            }
            switch (direction) {
                case Constants.DirDown:
                    this.moveTo(this.tileX, this.tileY + count);
                    break;
                case Constants.DirUp:
                    this.moveTo(this.tileX, this.tileY - count);
                    break;
                case Constants.DirLeft:
                    this.moveTo(this.tileX - count, this.tileY);
                    break;
                case Constants.DirRight:
                    this.moveTo(this.tileX + count, this.tileY);
                    break;
            }
        }

        out.canPass = mutable(function canPass(ent) {
            return !this.obstacle;
        });

        out.toString = function toString() {
            return "[object Entity(" + this.tileX + "," + this.tileY + ")]";
        }

        out.acting = mutable(property({
            get: function () { return this.moving; }
        }));

        out.hat = property({
            get: function () { return this._hat; },
            set: function (v) { this._hat = v; }
        });

        return out;
    }({}));
    ns.Entity = Entity;

    var Player = Class(Entity, function (out) {
        out.constructor = function Player(map, id, tileX, tileY) {
            Entity.call(this, map, id, tileX, tileY);
            this.isPlayer = true;
            this.keepMoving = true;
            this.animations = Animations.playerAnims;
            this.hatAnims = Animations.hatAnims;
        }

        return out;
    }({}));
    ns.Player = Player;

    var Guest = Class(Entity, function (out) {
        out.constructor = function Guest(map, id, type, tileX, tileY) {
            Entity.call(this, map, id, tileX, tileY);
            this.animations = Animations.guestAnims;
            this.hatAnims = Animations.guestHatAnims;
            this.hat = type;
            this.ko = false;
            this.koTurns = 0;

            var self = this;
            this.onStoppedMoving = function (e) { self.stoppedMoving.call(self, e); };
        }

        out.stoppedMoving = function stoppedMoving(self) {
            this.punchTargets();
        }

        out.isStrongAgainst = function isStrongAgainst(hat) {
            if (this.hat == 0) {
                return hat == 1;
            }
            else if (this.hat == 1) {
                return hat == 2;
            }
            else if (this.hat == 2) {
                return hat == 3;
            }
            else if (this.hat == 3) {
                return hat == 0;
            }
        }

        out.isWeakAgainst = function isWeakAgainst(hat) {
            if (this.hat == 0) {
                return hat == 3;
            }
            else if (this.hat == 1) {
                return hat == 0;
            }
            else if (this.hat == 2) {
                return hat == 1;
            }
            else if (this.hat == 3) {
                return hat == 2;
            }
        }

        out.slapFight = function slapFight(ent) {
            this.lock(this);
            this.lock(ent);
            if (this.tileX == ent.tileX) {
                this.map.addParticle(
                    new Particles.ExplosionEmitter(this.map, 0,
                        this.tileX,
                        Math.min(this.tileY, ent.tileY) + Math.abs(this.tileY - ent.tileY) / 2,
                        0.3, 1.3));
            }
            else {
                this.map.addParticle(
                    new Particles.ExplosionEmitter(this.map, 0,
                        Math.min(this.tileX, ent.tileX) + Math.abs(this.tileX - ent.tileX) / 2,
                        this.tileY,
                        1.3, 0.3));
            }
        }

        out.lock = function lock(ent) {
            ent.locked = true;
            ent.switchAnimation("ko"); // TODO: switch with looping angry anim (with punches?)
        }

        out.knockOut = function knockOut(ent) {
            this.punch(ent, "angry", "ko", false);
            ent.ko = true;
            ent.koTurns = 9;
            ent.koStars = new Particles.KoStars(ent.tileX, ent.tileY);
            ent.koCount = new Particles.KoCount(ent.tileX, ent.tileY);
            this.map.addParticle(ent.koStars);
            this.map.addParticle(ent.koCount);
        }

        out.knockBack = function knockBack(ent) {
            if (ent.ko) {
                this.punch(ent, "angry", "ko", true);
            }
            else {
                this.punch(ent, "angry", "hurt", true);
            }
        }

        out.punch = function punch(ent, thisAnim, thatAnim, push) {
            if (this.tileX > ent.tileX) {
                this.map.addParticle(new Particles.Punch(Engine.DirLeft, ent.tileX, ent.tileY));
                if (push && this.map.canMoveTo(ent, ent.tileX - 1, ent.tileY)) {
                    ent.tileX--;
                }
                ent.facing = Engine.DirRight;
            }
            else if (this.tileX < ent.tileX) {
                this.map.addParticle(new Particles.Punch(Engine.DirRight, ent.tileX, ent.tileY));
                if (push && this.map.canMoveTo(ent, ent.tileX + 1, ent.tileY)) {
                    ent.tileX++;
                }
                ent.facing = Engine.DirLeft;
            }
            else if (this.tileY > ent.tileY) {
                this.map.addParticle(new Particles.Punch(Engine.DirUp, ent.tileX, ent.tileY));
                if (push && this.map.canMoveTo(ent, ent.tileX, ent.tileY - 1)) {
                    ent.tileY--;
                }
                ent.facing = Engine.DirDown;
            }
            else if (push && this.tileY < ent.tileY) {
                this.map.addParticle(new Particles.Punch(Engine.DirDown, ent.tileX, ent.tileY));
                if (this.map.canMoveTo(ent, ent.tileX, ent.tileY + 1)) {
                    ent.tileY++;
                }
                ent.facing = Engine.DirUp;
            }
            ent.switchAnimation(thatAnim);
            this.switchAnimation(thisAnim);
        }

        out.takeTurn = function takeTurn(player) {
            if (this.ko || this.locked) {
                return;
            }
            this.punchTargets();
        }

        out.punchTargets = function punchTargets() {
            for (var i = 0; i < this.map.entities.count; i++) {
                var ent = this.map.entities.get(i);
                if ((Math.abs(ent.tileX - this.tileX) == 1 && ent.tileY - this.tileY == 0) ||
                    (Math.abs(ent.tileY - this.tileY) == 1 && ent.tileX - this.tileX == 0)) {
                    if (this.hat != ent.hat && !(ent instanceof PickUp)) {
                        if (ent.isPlayer) {
                            this.knockBack(ent);
                            this.map.pause(Constants.PunchPause);
                            this.map.Game.PlaySound("pow");
                        }
                        else if (!ent.ko && !ent.locked) {
                            if (this.isStrongAgainst(ent.hat)) {
                                console.writeLine("Giga drill paaaaanch!");
                                this.knockOut(ent);
                                this.map.pause(Constants.PunchPause);
                                this.map.Game.PlaySound("pow");
                            }
                            else if (this.isWeakAgainst(ent.hat)) {
                                console.writeLine("Meh punch. :<");
                                this.knockBack(ent);
                                this.map.pause(Constants.PunchPause);
                                this.map.Game.PlaySound("pow");
                            }
                            else {
                                console.writeLine("Omg slapfight!");
                                this.slapFight(ent);
                                this.map.pause(Constants.PunchPause);
                                this.map.Game.PlaySound("pow");
                            }
                        }
                    }
                }
            }
        }

        out.reduceKoCounter = function reduceKoCounter() {
            if (!this.ko) {
                return;
            }
            this.koTurns--;
            if (this.koTurns == 0) {
                this.ko = false;
                this.map.removeParticle(this.koStars);
                this.map.removeParticle(this.koCount);
                this.switchAnimation("stand");
                this.punchTargets();
            }
            else {
                this.koCount.setCounter(this.koTurns);
            }
        }

        out.canPass = mutable(function canPass(ent) {
            return this.ko || base(this).canPass.call(this, ent);
        });

        out.acting = mutable(property({
            get: function () {
                return this.moving || this.curAnim == "angry";
            }
        }));

        return out;
    }({}));
    ns.Guest = Guest;

    var PickUp = Class(Entity, function (out) {
        out.constructor = function PickUp(map, id, tileX, tileY) {
            Entity.call(this, map, id, tileX, tileY);
            this.isPickUp = true;
            this.obstacle = false;
        }

        out.canPass = mutable(function canPass(ent) {
            if (ent.isPlayer) {
                this.pickedUp(ent);
                this.map.removeEntity(this);
            }
            return true;
        });

        out.pickedUp = function pickedUp(ent) {
        }

        return out;
    }({}));
    ns.PickUp = PickUp;

    var Key = Class(PickUp, function (out) {
        out.constructor = function Key(map, id, tileX, tileY) {
            PickUp.call(this, map, id, tileX, tileY);
            this.animations = Animations.keyAnims;
        }

        out.pickedUp = function pickedUp(ent) {
            if (ent.isPlayer) {
                ent.hasKey = true;
                this.map.Hud.hasKey = true;
                console.writeLine("Picked up key");
                this.map.Game.PlaySound("kagi1");
            }
        }

        return out;
    }({}));
    ns.Key = Key;

    var Hat = Class(PickUp, function (out) {
        out.constructor = function Hat(map, id, type, tileX, tileY) {
            PickUp.call(this, map, id, tileX, tileY);
            this.type = type;
            this.animations = Animations.hatItemAnims[type];
        }

        out.pickedUp = function pickedUp(ent) {
            if (ent.isPlayer) {
                console.writeLine("Picked up hat");
                ent.hat = this.type;
                this.map.Game.PlaySound("hat3");
            }
        }

        return out;
    }({}));
    ns.Hat = Hat;

    var HatA = Class(Hat, function (out) {
        out.constructor = function HatA(map, id, tileX, tileY) {
            Hat.call(this, map, id, 0, tileX, tileY);
        }

        return out;
    }({}));
    ns.HatA = HatA;

    var HatB = Class(Hat, function (out) {
        out.constructor = function HatB(map, id, tileX, tileY) {
            Hat.call(this, map, id, 1, tileX, tileY);
        }

        return out;
    }({}));
    ns.HatB = HatB;

    var HatC = Class(Hat, function (out) {
        out.constructor = function HatC(map, id, tileX, tileY) {
            Hat.call(this, map, id, 2, tileX, tileY);
        }

        return out;
    }({}));
    ns.HatC = HatC;

    var HatD = Class(Hat, function (out) {
        out.constructor = function HatD(map, id, tileX, tileY) {
            Hat.call(this, map, id, 3, tileX, tileY);
        }

        return out;
    }({}));
    ns.HatD = HatD;

    var Door = Class(Entity, function (out) {
        out.constructor = function Door(map, id, tileX, tileY) {
            Entity.call(this, map, id, tileX, tileY);
            this.animations = Animations.doorAnims;
        }

        out.canPass = mutable(function canPass(ent) {
            if (this.isOpen) {
                return true;
            }
            if (ent.hasKey) {
                this.map.Game.PlaySound("kagi4");
                this.switchAnimation("open");
                this.isOpen = true;
                console.writeLine("Door opened. You did it!");
                this.map.finish();
                return true;
            }
            return false;
        });

        return out;
    }({}));
    ns.Door = Door;

    return ns;
}());